#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2020-04-22
#
# inventory of Cisco BGP Peers (IPv4 and IPv6)
#
# 2022-04-22: moved here from cisco_bgp_peer
# 2022-04-30: code cleanup/streamlining
#

from typing import List
from cmk.base.plugins.agent_based.agent_based_api.v1 import (
    register,
    SNMPTree,
    exists,
    contains,
    all_of,
    OIDEnd,
    OIDBytes,
)
from cmk.base.plugins.agent_based.agent_based_api.v1.type_defs import (
    StringByteTable
)
from cmk.base.plugins.agent_based.utils.bgp_peer import (
    InvBgpPeer,
    bgp_error_as_string,
    bgp_render_ip_address,
    bgp_get_ip_address_from_oid,
    bgp_error_code_as_hex,
    get_bgp_type,
)


def parse_inv_cisco_bgp_peer(string_table: List[StringByteTable]):
    peers, address_family, base = string_table
    try:
        base_local_as = base[0][0]
    except (ValueError, IndexError):
        base_local_as = 'N/A'

    bgp_peers = []

    peer_families = dict(address_family)

    for entry in peers:
        try:
            oid_end, state, local_addr, local_as, local_id, remote_as, remote_id, last_error, \
            fsm_established_time, last_errortxt = entry
        except ValueError:
            return

        if len(local_addr) == 4:
            local_addr_type = '1'  # IPv4
        elif len(local_addr) == 16:
            local_addr_type = '2'  # IPv6
        else:
            local_addr_type = '0'

        if local_as == '0':
            local_as = base_local_as

        address_families = []
        for key in peer_families:
            if key.startswith(oid_end):
                address_families.append(peer_families[key])

        bgp_peer: InvBgpPeer = {
            'remote_addr': bgp_get_ip_address_from_oid(oid_end),
            'remote_id': remote_id,
            'version': '4',
            'local_addr': bgp_render_ip_address(local_addr_type, local_addr),
            'remote_as': remote_as,
            'local_as': local_as,
            'local_id': local_id,
            'bgp_type': get_bgp_type(local_as, remote_as),
            'fsm_established_time': int(fsm_established_time) if fsm_established_time.isdigit() else None,
            'peer_state': 1 if state == '6' else 2,  # adjust to match if_oper_status for inventory painter
            'last_error_code': bgp_error_code_as_hex(last_error),
            'last_error': bgp_error_as_string(last_error),
            'status_columns': {},
            'address_family': ','.join(address_families)
        }

        bgp_peers.append(bgp_peer)

    return bgp_peers


register.snmp_section(
    name='inv_cisco_bgp_peer',
    parse_function=parse_inv_cisco_bgp_peer,
    parsed_section_name='inv_bgp_peer',
    supersedes=['inv_bgp_peer'],
    fetch=[
        SNMPTree(
            base='.1.3.6.1.4.1.9.9.187.1.2.5.1',  # CISCO-BGP4-MIB::cbgpPeer2Entry
            oids=[
                OIDEnd(),
                '3',  # cbgpPeer2State
                OIDBytes('6'),  # cbgpPeer2LocalAddr
                '8',  # cbgpPeer2LocalAs -> empty
                '9',  # cbgpPeer2LocalIdentifier
                '11',  # cbgpPeer2RemoteAs
                '12',  # cbgpPeer2RemoteIdentifier
                OIDBytes('17'),  # cbgpPeer2LastError
                '19',  # cbgpPeer2FsmEstablishedTime
                '28',  # cbgpPeer2LastErrorTxt
            ]
        ),
        SNMPTree(
            base='.1.3.6.1.4.1.9.9.187.1.2',  # cbgpPeer
            oids=[
                OIDEnd(),  #
                '7.1.3',  # cbgpPeer2AddrFamilyName
            ]
        ),
        SNMPTree(
            base='.1.3.6.1.2.1.15',  # BGP-4-MIB
            oids=[
                '2',  # bgpLocalAs
            ]
        )
    ],
    detect=all_of(
        contains('.1.3.6.1.2.1.1.1.0', 'Cisco'),
        exists('.1.3.6.1.4.1.9.9.187.1.2.5.1.3.*')
    ),
)
